package require Expect

# create the abc namespace if necessary
namespace eval abc {}

# disable Expect logging
log_user 0

# include configuration file
# relative scripts JS
source [file join [file dirname [info script]] abctcl.conf.tcl]
# lock variable
# all implemented commands set it to 1 when communicating with ABC - in this
# state all other commands called from other processes (especially in GUI) are
# refused to communicate with ABC
# variable is set to 0 when communication completed
set abc::busy 0
# notices
set abc::messages(ABC_N_EQ_YES)  {The networks are equivalent.}
set abc::messages(ABC_N_SAT_YES) {The formula is satisfiable.}
set abc::messages(ABC_N_EQ_NO)   {The networks are not equivalent.}
set abc::messages(ABC_N_SAT_NO)  {The formula is unsatisfiable.}
set abc::messages(ABC_N_EQ_UND)  {The equivalence of the networks could not be determined.}
set abc::messages(ABC_N_SAT_UND) {The satisfiablility of the formula could not be determined.}
# file errors
set abc::messages(ABC_E_READ)    {Error reading %s: %s.}
set abc::messages(ABC_E_EXEC)    {Error executing %s.} 
set abc::messages(ABC_E_SUFF)    {Unknown file suffix %s.}
set abc::messages(ABC_E_OPEN)    {Error opening file %s.}
set abc::messages(ABC_E_EXF)     {%s does not exist or is not executable.} 
set abc::messages(ABC_E_EXD)     {%s does not exist or is not a directory.} 
# ABC errors
set abc::messages(ABC_E_PARAMS)  {Incorrect parameters}       ;# elaborate
set abc::messages(ABC_E_CMD)     {ABC command error: %s}
set abc::messages(ABC_E_EMPTY)   {An empty network detected by ABC.}
set abc::messages(ABC_E_PARSE)   {The output could not be parsed.}
set abc::messages(ABC_E_INCOMP)  {Compared networks differ in inputs and/or outputs.}
set abc::messages(ABC_E_TOUT)    {ABC command timeout.}       ;#%d
set abc::messages(ABC_E_EOF)     {Unexpected end-of-file from ABC.}
set abc::messages(ABC_E_BUSY)    {Another ABC command in progress.}

proc abc::errorReturn {resArrayName code args} {
    upvar $resArrayName res;
    set code [abc::consCode $code {*}$args]
    set res(errorCode) $code
    return -level 2 -code error -errorcode $code -errorinfo "error: $code"
}

proc abc::consCode {code args} {
    global abc::messages
    return [concat $code [format $abc::messages($code) {*}$args]]
}

proc abc::extrSym {code} {
    return [lindex $code 0]
}

proc abc::extrMesg {code} {
    return [lrange $code 1 end]
}

proc abc::extrSeverity {code} {
    set symbol [lindex $code 0];    # does not hurt if only a symbol is given
    if { [regexp {ABC_(.)_} $symbol match v1] } { return $v1 }
}

# executes ABC program, waits for first prompt, sources abc.rc
# expects prompt as output from ABC, waits for only first occurence, prompt looks
# like this: 'abc 123> '
# waits maximal 'timeout' seconds - set in settings file
# no. of bytes returned in buffer is maximaly abc::expect_buffer_size - set in settings file
# resultArrayName is optional and contains fields:
#     spawn_id      [int]                 id returned from 'spawn'
#     rawOutput     [string]              ABC's total output including prompt
#     errorCode     [string]              error code on return; NONE if ok
#     promptNumber  [int]                 prompt number
#     cmdTime       [int]                 time in milliseconds
#     abcVersion    [string]              initial ABC's announcement
#     sourced       [list of strings]     initialization files sourced into ABC
#     failed        [list of strings]     initialization files that failed to source into ABC
#
# possible error codes: 
# ABC_E_EXD ABC_E_EXF ABC_E_EXEC ABC_E_CMD ABC_E_BUSY ABC_E_EOF ABC_E_TOUT
#

proc abc::init { { resultArrayName {}} {expectTimeout {}}} {
	if {$resultArrayName != {}} {
		upvar $resultArrayName res
	}
	global abc::dir abc::exe abc::initList
	global spawn_id
	global errorCode errorInfo
    global timeout abc::busy
	global abc::expect_buffer_size

	set res(promptNumber) null
	set res(rawOutput) null
	set res(spawn_id) null
	set res(sourced) {}
	set res(failed) {}
	set res(cmdTime) 0
	set res(abcVersion) null

    set abcAbsExe [file join $abc::dir $abc::exe]

	# test whether $abcDir is directory
	if {[file isdirectory $abc::dir] == 0} {
        abc::errorReturn res {ABC_E_EXD} $abcDir
	}
	# test whether $abcExe file exists (file executable??)
	if {[file executable $abcAbsExe ] == 0} {
        abc::errorReturn res {ABC_E_EXF} $abcAbsExe
	}

	# program execution
	set res(spawn_id) [spawn $abcAbsExe "-s"]

	if {$res(spawn_id) == 0} { 
        abc::errorReturn res {ABC_E_EXEC} $abcAbsExe
	}
	# process successfully spawned - waiting for first prompt
	set saveTimeout $timeout 
    if {$expectTimeout != {}} { set timeout $expectTimeout }
	
	match_max $abc::expect_buffer_size

	expect {
		-re {abc ([0-9]+)> } {
			# prompt succesfully catched
			set res(rawOutput) $expect_out(buffer)
			set res(promptNumber) $expect_out(1,string)
			regexp {^([^\r\n]*)} $expect_out(buffer) match res(abcVersion) 
		}
		eof {
		    set abc::busy 0
            set timeout $saveTimeout 
		    abc::errorReturn res {ABC_E_EOF}
		}
		timeout {
		    set abc::busy 0   ;# ?????
            set timeout $saveTimeout 
			abc::errorReturn res {ABC_E_TOUT} 
		}
	}
    set abc::busy 0
    set timeout $saveTimeout 

	# init files force-fed into ABC
	foreach initFile $abc::initList {
        set rtn [catch { abc::source $initFile sourceResult } sourceReturn sourceDict]
		catch { set res(cmdTime) [expr {$res(cmdTime) + $sourceResult(cmdTime)} ] }
        catch { set res(promptNumber) $sourceResult(promptNumber) } 
        catch { append res(rawOutput) $sourceResult(rawOutput) } 
        if {$rtn == 0} {
		    lappend res(sourced) $initFile
            continue
        }
        if {[abc::extrSym $sourceResult(errorCode)] == {ABC_E_READ}} {
            lappend res(failed) $initFile
            continue
        }
        return -options $sourceDict;        # rethrow
	}
    return -code ok
}

# kills ABC process, closes communication channel
proc abc::done {} {
	# alias for send
	exp_send "quit\r"
	close
}

# send command to ABC and wait for prompt
# resultArrayName optional return parameter - an array which contains following
# fields (key, value)
# keys:   contains 'null' when no prompt catched:
#         errorCode     [string]   error message, NONE if OK
#         rawOutput     [string]   the entire output of ABC including prompt
#         promptNumber  [int]      prompt number
#         cmdTime       [int]      time in milliseconds
#
# errors: ABC_E_EOF ABC_E_TOUT ABC_E_BUSY
#
proc abc::command {command {resultArrayName {}} {expectTimeout {}}} {
    global timeout abc::busy
	global abc::expect_buffer_size

	if {$resultArrayName != {}} { upvar $resultArrayName res }

	set res(errorCode) "NONE"   ;# as the system errorCode does
	set res(rawOutput) null
	set res(promptNumber) null

    # check ABC status
	if {$abc::busy == 1} { abc::errorReturn res {ABC_E_BUSY} } 
	set abc::busy 1

    # if specified, set Expect's timeout 
    set saveTimeout $timeout 
    if {$expectTimeout != {}} { set timeout $expectTimeout }

    # send the command to ABC
	exp_send "$command\r"
	
	# accept output till the prompt is recognized
	set res(promptNumber) null

	match_max $abc::expect_buffer_size
	
	set timeStart [clock clicks -milliseconds]
	expect {
		-re {abc ([0-9]+)> } {
			# prompt succesfully catched
			set res(rawOutput) $expect_out(buffer)
			set res(promptNumber) $expect_out(1,string)
		}
		eof {
		    set abc::busy 0
            set timeout $saveTimeout 
		    abc::errorReturn res {ABC_E_EOF}
		}
		timeout {
		    set abc::busy 0   ;# ?????
            set timeout $saveTimeout 
			abc::errorReturn res {ABC_E_TOUT}
		}
	}
	set timeStop [clock clicks -milliseconds]
    set res(cmdTime) [expr {$timeStop - $timeStart}]
    set abc::busy 0
    
    # restore timeout
    set timeout $saveTimeout 

	return -code ok
}
#
# source a file into ABC, parse errors
# file errors are <filename>: <system error msg>
# interpretation errors are ** cmd error: <abc error msg>
# resultArrayName optional return parameter - an array which contains following
# fields (key, value)
# keys:   contains 'null' when no prompt catched:
#         errorCode     [string]   error message, NONE if OK
#         rawOutput     [string]   the entire output of ABC including prompt
#         promptNumber  [int]      prompt number
#         cmdTime       [int]      time in milliseconds
#
# errors: ABC_E_EOF ABC_E_TOUT ABC_E_BUSY ABC_E_CMD ABC_E_READ
#
proc abc::source {filename {resultArrayName {}} {expectTimeout {}}} {

	if {$resultArrayName != {}} { upvar $resultArrayName res }

    abc::command "source $filename" res $expectTimeout
	# no need to catch exceptional result from abc::command
    # we are still here so the return must have been ok
    
    set cmdMatch [regexp {^\*\* cmd error: ([^\r\n]*)} $res(rawOutput) match matchErr]
    if { $cmdMatch } { 
        abc::errorReturn res {ABC_E_CMD}  $matchErr
    }
    
    set cmdMatch [regexp {^.*: ([^\r\n]*)} $res(rawOutput) match matchErr]
    if { $cmdMatch } { 
        abc::errorReturn res {ABC_E_READ} $filename $matchErr
    }
    # if the output size is not zero then report unparseable output
    return -code ok
}

# send print_stats command to ABC
# resultArrayName optional return parameter - an array which contains following
# fields (key, value)
# keys:   contains 'null' when value was not recognized:
#         i            [int]
#         o            [int]
#         a            [int]
#         lat          [int]
#         lev          [int]
#         edge         [int]
#         net          [int]
#         nd           [int]
#         wbox         [int]
#         bbox         [int]
#         choice       [int]
#         exor         [int]
#         mux          [int]
#         pure_and     [int]
#         and          [int]
#         cube         [int]
#         aig          [int]
#         lit_fac      [int]
#         area         [double]
#         delay        [double]
#         power        [double]
#         glitch       [double]
#         errorCode    [string] error message, NONE if OK
#         rawOutput    [string]
#         promptNumber [int]    prompt number
#         cmdTime      [int]    time in milliseconds
# 
# paramsString - optional string containing parameters for print_stats command,
# string is passed onto print_stats 'as is'
# return: ok or error
# abc90813_source\src\base\abci\abcPrint.c
#
#   errors: ABC_E_EOF, ABC_E_TOUT  ABC_E_BUSY ABC_E_PARSE ABC_E_EMPTY
#
proc abc::print_stats { {resultArrayName {}}  {paramsString {}} {expectTimeout {}}} {
	global abc::busy

	if {$resultArrayName != {}} {
		upvar $resultArrayName res
	}

    foreach item {
	i o	a lat lev edge net nd wbox bbox choice exor mux pure_and and cube aig lit_fac area delay power glitch
	} { set {$res($item)} null }

    abc::command "print_stats $paramsString" res $expectTimeout

	set matched 0

	# parse output - apply a series of regular expressions to the output string
	set regexpRes [regexp -- {[[:blank:]]+(i)/(o) =[[:blank:]]*([0-9]+)/[[:blank:]]*([0-9]+)} $res(rawOutput) match v1 v2 v3 v4]
	if {$regexpRes} {
		set res($v1) $v3
		set res($v2) $v4
		incr matched
	}

	set regexpRes [regexp -- {[[:blank:]]+(i)/(o)/(a) =[[:blank:]]*([0-9]+)/[[:blank:]]*([0-9]+)/[[:blank:]]*([0-9]+)} $res(rawOutput) match v1 v2 v3 v4 v5 v6]
	if {$regexpRes} {
		set res($v1) $v4
		set res($v2) $v5
		set res($v3) $v6
		incr matched
	}

	set regexpRes [regexp -- {[[:blank:]]+(lat) =[[:blank:]]*([0-9]+)[[:blank:]]*} $res(rawOutput) match v1 v2]
	if {$regexpRes} {
		set res($v1) $v2
		incr matched
	}

	set regexpRes [regexp -- {[[:blank:]]+(lev) =[[:blank:]]*([0-9]+)[[:blank:]]*} $res(rawOutput) match v1 v2]
	if {$regexpRes} {
		set res($v1) $v2
		incr matched
	}

	set regexpRes [regexp -- {[[:blank:]]+(edge) =[[:blank:]]*([0-9]+)[[:blank:]]*} $res(rawOutput) match v1 v2]
	if {$regexpRes} {
		set res($v1) $v2
		incr matched
	}

	if { [regexp -- {[[:blank:]]+(net) =[[:blank:]]*([0-9]+)[[:blank:]]*} $res(rawOutput) match v1 v2] } {
		set res($v1) $v2
		incr matched
	}

	if { [regexp -- {[[:blank:]]+(nd) =[[:blank:]]*([0-9]+)[[:blank:]]*} $res(rawOutput) match v1 v2] } {
		set res($v1) $v2
		incr matched
	}

	if { [regexp -- {[[:blank:]]+(wbox) =[[:blank:]]*([0-9]+)[[:blank:]]*} $res(rawOutput) match v1 v2] } {
		set res($v1) $v2
		incr matched
	}

	if { [regexp -- {[[:blank:]]+(bbox) =[[:blank:]]*([0-9]+)[[:blank:]]*} $res(rawOutput) match v1 v2] } {
		set res($v1) $v2
		incr matched
	}

	if { [regexp -- {\((choice) =[[:blank:]]*([0-9]+)\)} $res(rawOutput) match v1 v2] } {
		set res($v1) $v2
		incr matched
	}

	if { [regexp -- {\((exor) =[[:blank:]]*([0-9]+)\)} $res(rawOutput) match v1 v2] } {
		set res($v1) $v2
		incr matched
	}

	if { [regexp -- {\((mux) =[[:blank:]]*([0-9]+)\)} $res(rawOutput) match v1 v2] } {
		set res($v1) $v2
		incr matched
	}

	if { [regexp -- {\(pure and =[[:blank:]]*([0-9]+)\)} $res(rawOutput) match v1] } {
		set res(pure_and) $v1
		incr matched
	}

	set regexpRes [regexp -- {[[:blank:]]+(and) =[[:blank:]]*([0-9]+)[[:blank:]]*} $res(rawOutput) match v1 v2]
	if {$regexpRes} {
		set res($v1) $v2
		incr matched
	}

	set regexpRes [regexp -- {[[:blank:]]+(cube) =[[:blank:]]*([0-9]+)[[:blank:]]*} $res(rawOutput) match v1 v2]
	if {$regexpRes} {
		set res($v1) $v2
		incr matched
	}
	if { [regexp -- {[[:blank:]]+lit\(fac\) =[[:blank:]]*([0-9]+)[[:blank:]]+} $res(rawOutput) match v1] } {
		set res(lit_fac) $v1
		incr matched
	}

	set regexpRes [regexp -- {[[:blank:]]+(aig)[[:blank:]]*=[[:blank:]]*([0-9]+)[[:blank:]]*} $res(rawOutput) match v1 v2]
	if {$regexpRes} {
		set res($v1) $v2
		incr matched
	}

	if { [regexp -- {[[:blank:]]+(bdd)[[:blank:]]*=[[:blank:]]*([0-9]+)[[:blank:]]*} $res(rawOutput) match v1 v2] } {
		set res($v1) $v2
		incr matched
	}

	if { [regexp -- {[[:blank:]]*(area) =[[:blank:]]*([0-9]+.[0-9]+)[[:blank:]]} $res(rawOutput) match v1 v2] } {
		set res($v1) $v2
		incr matched
	}

	if { [regexp -- {[[:blank:]]*(delay) =[[:blank:]]*([0-9]+.[0-9]+)[[:blank:]]} $res(rawOutput) match v1 v2] } {
		set res($v1) $v2
		incr matched
	}

	if { [regexp -- {[[:blank:]]*(power) =[[:blank:]]*([0-9]+.[0-9]+)[[:blank:]]} $res(rawOutput) match v1 v2] } {
		set res($v1) $v2
	}

	if { [regexp -- {[[:blank:]]*(glitch) =[[:blank:]]*([0-9]+.[0-9]+)[[:blank:]]} $res(rawOutput) match v1 v2] } {
		set res($v1) $v2
		incr matched
	}

    
	if {[regexp -- {Empty network} $res(rawOutput)]} {
		abc::errorReturn res {ABC_E_EMPTY}
	}

    if { $matched == 0 } {
		abc::errorReturn res {ABC_E_PARSE}
    }
    return -code ok
}
# executes ABC's 'print_latch' command
# resultArrayName is an optional parameter, consists of pairs (key, value):
# keys: contains 'null' when value was not recognized:
#       total_latches    [int]
#       init0            [int]
#       init1            [int]
#       initDC           [int]
#       const_data       [int]
#       errorCode        [string] error message, NONE if OK
#       rawOutput        [string]
#       promptNumber     [int]    prompt number
#       cmdTime          [int]    time in milliseconds
# paramsString - optional string containing parameters for print_latch command,
# string is passed onto print_latch 'as is'
# return: ok or error
#
#   errors: ABC_E_EOF, ABC_E_TOUT  ABC_E_BUSY ABC_E_PARSE ABC_E_EMPTY
#
# abc 02> print_latch
# The network is combinational. 
# Covered by 'null'
proc abc::print_latch { {resultArrayName {}} {paramsString {}} {expectTimeout {}}} {
	global abc::busy

	if {$resultArrayName != {}} {
		upvar $resultArrayName res
	}
	set res(total_latches) null
	set res(init0) null
	set res(init1) null
	set res(initDC) null
	set res(const_data) null

    abc::command "print_latch $paramsString" res $expectTimeout

	set matched 0
    #
    #   The original set by Petr
    #
	if {[regexp -- {Total latches =[[:blank:]]*([0-9]+)} $res(rawOutput) match v1]} {
		set res(total_latches) $v1
		incr matched
	}
	if {[regexp -- {Init0 =[[:blank:]]*([0-9]+)} $res(rawOutput) match v1]} {
		set res(init0) $v1
		incr matched
	}
	if {[regexp -- {Init1 =[[:blank:]]*([0-9]+)} $res(rawOutput) match v1]} {
		set res(init1) $v1
		incr matched
	}
	if {[regexp -- {InitDC =[[:blank:]]*([0-9]+)} $res(rawOutput) match v1]} {
		set res(initDC) $v1
		incr matched
	}
	if {[regexp -- {Const data =[[:blank:]]*([0-9]+)} $res(rawOutput) match v1]} {
		set res(const_data) $v1
		incr matched
	}
	if {[regexp -- {Empty network} $res(rawOutput)]} {
		abc::errorReturn res {ABC_E_EMPTY}
	}
	#
	#   ABC 7
	#
	if {[regexp -- {Latch =[[:blank:]]*([0-9]+)} $res(rawOutput) match v1]} {
		set res(total_latches) $v1
		incr matched
	}
	if {[regexp -- {Zero =[[:blank:]]*([0-9]+)} $res(rawOutput) match v1]} {
		set res(init0) $v1
		incr matched
	}
	if {[regexp -- {One =[[:blank:]]*([0-9]+)} $res(rawOutput) match v1]} {
		set res(init1) $v1
		incr matched
	}
	if {[regexp -- {DC init =[[:blank:]]*([0-9]+)} $res(rawOutput) match v1]} {
		set res(initDC) $v1
		incr matched
	}
	if {[regexp -- {Const fanin =[[:blank:]]*([0-9]+)} $res(rawOutput) match v1]} {
		set res(const_data) $v1
		incr matched
	}
	if {[regexp -- {Empty network} $res(rawOutput)]} {
		abc::errorReturn res {ABC_E_EMPTY}
	}
	if {[regexp -- {The network is combinational} $res(rawOutput)]} {
    	set res(total_latches) 0
    	set res(init0) 0
    	set res(init1) 0
    	set res(initDC) 0
    	set res(const_data) 0
    	incr matched
	}
    if { $matched == 0 } {
		abc::errorReturn res {ABC_E_PARSE}
    }
	return -code ok
}
# executes ABC's 'print_auto' command
# parses only cumulative statistics, ToDo: parse Output-Input-AutoK table
# resultArrayName is an optional parameter, consists of pairs (key, value):
# keys: contains 'null' when value was not recognized:
#       ins           [int]
#       inMax         [int]
#       outs          [int]
#       auto          [int]
#       sumK          [int]
#       kMax          [int]
#       supp          [int]
#       time          [float]
#       errorCode     [string] error message, NONE if OK
#       rawOutput     [string]
#       promptNumber  [int]    prompt number
#       cmdTime       [int]    time in milliseconds, real time
#
# paramsString - optional string containing parameters for print_auto command,
# string is passed onto print_auto 'as is'
# return: ok or error
#
#   errors: ABC_E_EOF, ABC_E_TOUT  ABC_E_BUSY ABC_E_PARSE ABC_E_EMPTY
#
# abc 02> print_auto
# This command works only for AIGs (run "strash").
# not covered.
proc abc::print_auto { {resultArrayName {}} {paramsString {}} {expectTimeout {}}} {

	if {$resultArrayName != {}} {
		upvar $resultArrayName res
	}

	set res(ins) null
	set res(inMax) null
	set res(outs) null
	set res(auto) null
	set res(sumK) null
	set res(kMax) null
	set res(supp) null
	set res(time) null
	
    abc::command "print_auto $paramsString" res $expectTimeout

	set matched 0

	if {[regexp -- {Ins=[[:blank:]]*([0-9]+)} $res(rawOutput) match v1]} {
		set res(ins) $v1
		incr matched
	}
	if {[regexp -- {InMax=[[:blank:]]*([0-9]+)} $res(rawOutput) match v1]} {
		set res(inMax) $v1
		incr matched
	}
	if {[regexp -- {Outs=[[:blank:]]*([0-9]+)} $res(rawOutput) match v1]} {
		set res(outs) $v1
		incr matched
	}
	if {[regexp -- {Auto=[[:blank:]]*([0-9]+)} $res(rawOutput) match v1]} {
		set res(auto) $v1
		incr matched
	}
	if {[regexp -- {SumK=[[:blank:]]*([0-9]+)} $res(rawOutput) match v1]} {
		set res(sumK) $v1
		incr matched
	}
	if {[regexp -- {KMax=[[:blank:]]*([0-9]+)} $res(rawOutput) match v1]} {
		set res(kMax) $v1
		incr matched
	}
	if {[regexp -- {Supp=[[:blank:]]*([0-9]+)} $res(rawOutput) match v1]} {
		set res(supp) $v1
		incr matched
	}
	if {[regexp -- {Time=[[:blank:]]*([0-9]+.[0-9]+)} $res(rawOutput) match v1]} {
		set res(time) $v1
		incr matched
	}
	if {[regexp -- {Empty network} $res(rawOutput)]} {
		abc::errorReturn res {ABC_E_EMPTY}
	}
    if { $matched == 0 } {
		abc::errorReturn res {ABC_E_PARSE}
    }
	return -code ok
}
# executes ABC's 'print_fanio' command
# parses only fanins/fanouts statistics, ToDo: parse Nodes table
# resultArrayName is an optional parameter, consists of pairs (key, value):
# keys: contains 'null' when value was not recognized:
#       fanins_max   [int]
#       fanins_ave   [float]
#       fanouts_max  [int]
#       fanouts_ave  [float]
#       errorCode    [string] error message, NONE if OK
#       rawOutput    [string]
#       promptNumber [int]    prompt number
#       cmdTime      [int]    time in milliseconds
#
# paramsString - optional string containing parameters for print_fanio command,
# string is passed onto print_fanio 'as is'
# return: ok or error
#
#   errors: ABC_E_EOF, ABC_E_TOUT  ABC_E_BUSY ABC_E_PARSE ABC_E_EMPTY
#
# BUG (ABC 70930): output not parsed
# The distribution of fanins and fanouts in the network:
#  Number   Nodes with fanin  Nodes with fanout
#    1 :            8               131  
#    2 :          155                23  
# 
proc abc::print_fanio { {resultArrayName {}} {paramsString {}} {expectTimeout {}}} {
	if {$resultArrayName != {}} {
		upvar $resultArrayName res
	}
	set res(fanins_max) null
	set res(fanins_ave) null
	set res(fanouts_max) null
	set res(fanouts_ave) null

    abc::command "print_fanio $paramsString" res $expectTimeout

	set matched 0

	if {[regexp -- {Fanins: Max = ([0-9]+). Ave = ([0-9]*.[0-9]*).  Fanouts: Max = ([0-9]+). Ave =  ([0-9]*.[0-9]*).} $res(rawOutput) match v1 v2 v3 v4]} {
		set res(fanins_max) $v1
		set res(fanins_ave) $v2
		set res(fanouts_max) $v3
		set res(fanouts_ave) $v4
		incr matched
	}

	if {[regexp -- {Empty network} $res(rawOutput)]} {
		abc::errorReturn res {ABC_E_EMPTY}
	}
    if { $matched == 0 } {
		abc::errorReturn res {ABC_E_PARSE}
    }
	return -code ok
}

# executes ABC's 'print_gates' command
# resultArrayName is an optional parameter, consists of pairs (key, value):
# keys: contains 'null' when value was not recognized:
#       const             [int]
#       const_percent     [float]
#       buffer            [int]
#       buffer_percent    [float]
#       inverter          [int]
#       inverter_percent  [float]
#       and               [int]
#       and_percent       [float]
#       or                [int]
#       or_percent        [float]
#       other             [int]
#       other_percent     [float]
#       total             [int]
#       total_percent     [float]
#       errorCode         [string] error message, NONE if OK
#       rawOutput         [string]
#       promptNumber      [int]    prompt number
#       cmdTime           [int]    time in milliseconds
#
# paramsString - optional string containing parameters for print_gates command,
# string is passed onto print_gates 'as is'
# return: ok or error
#
#   errors: ABC_E_EOF, ABC_E_TOUT  ABC_E_BUSY ABC_E_PARSE ABC_E_EMPTY
#
proc abc::print_gates { {resultArrayName {}} {paramsString {}} {expectTimeout {}}} {
	if {$resultArrayName != {}} {
		upvar $resultArrayName res
	}

	set res(const) null
	set res(const_percent) null
	set res(buffer) null
	set res(buffer_percent) null
	set res(inverter) null
	set res(inverter_percent) null
	set res(and) null
	set res(and_percent) null
	set res(or) null
	set res(or_percent) null
	set res(other) null
	set res(other_percent) null
	set res(total) null
	set res(total_percent) null

    abc::command "print_gates $paramsString" res $expectTimeout

	set matched 0

	if {[regexp -- {Const[[:blank:]]*=[[:blank:]]*([0-9]+)[[:blank:]]+([0-9]+.[0-9]+) %} $res(rawOutput) match v1 v2]} {
		set res(const) $v1
		set res(const_percent) $v2
		incr matched
	}
	if {[regexp -- {Buffer[[:blank:]]*=[[:blank:]]*([0-9]+)[[:blank:]]+([0-9]+.[0-9]+) %} $res(rawOutput) match v1 v2]} {
		set res(buffer) $v1
		set res(buffer_percent) $v2
		incr matched
	}
	if {[regexp -- {Inverter[[:blank:]]*=[[:blank:]]*([0-9]+)[[:blank:]]+([0-9]+.[0-9]+) %} $res(rawOutput) match v1 v2]} {
		set res(inverter) $v1
		set res(inverter_percent) $v2
		incr matched
	}
	if {[regexp -- {And[[:blank:]]*=[[:blank:]]*([0-9]+)[[:blank:]]+([0-9]+.[0-9]+) %} $res(rawOutput) match v1 v2]} {
		set res(and) $v1
		set res(and_percent) $v2
		incr matched
	}
	if {[regexp -- {Or[[:blank:]]*=[[:blank:]]*([0-9]+)[[:blank:]]+([0-9]+.[0-9]+) %} $res(rawOutput) match v1 v2]} {
		set res(or) $v1
		set res(or_percent) $v2
		incr matched
	}
	if {[regexp -- {Other[[:blank:]]*=[[:blank:]]*([0-9]+)[[:blank:]]+([0-9]+.[0-9]+) %} $res(rawOutput) match v1 v2]} {
		set res(other) $v1
		set res(other_percent) $v2
		incr matched
	}
	if {[regexp -- {TOTAL[[:blank:]]*=[[:blank:]]*([0-9]+)[[:blank:]]+([0-9]+.[0-9]+) %} $res(rawOutput) match v1 v2]} {
		set res(total) $v1
		set res(total_percent) $v2
		incr matched
	}

	if {[regexp -- {Empty network} $res(rawOutput)]} {
		abc::errorReturn res {ABC_E_EMPTY}
	}
    if { $matched == 0 } {
		abc::errorReturn res {ABC_E_PARSE}
    }
	return -code ok
}


# executes ABC's 'print_symm' command
# resultArrayName is an optional parameter, consists of pairs (key, value):
#
# keys: contains 'null' when value was not recognized:
#       some fields in output have same names - only 6 out of 23 fields were
#       implemented
#       total_func_supps   [int]
#       total_struct_supps [int]
#       total_symm         [int]
#       structural_symm    [int]
#       total_non_sym      [int]
#       total_var_pairs    [int]
#       errorCode          [string] error message, NONE if OK
#       rawOutput          [string]
#       promptNumber       [int]    prompt number
#       cmdTime            [int]    time in milliseconds
#
# REMARK (JS)
# The ABC output has two parts - on supports and on symmetries.
# Some entries in the parts do have equal names. In this case, 
# blind regexp does not suffice.
#
# paramsString - optional string containing parameters for print_symm command,
# string is passed onto print_symm 'as is'
# return: empty string
#
#   errors: ABC_E_EOF, ABC_E_TOUT  ABC_E_BUSY ABC_E_PARSE ABC_E_EMPTY
#
# BUG: message not detected
# This command works only for combinational networks (run "comb").
proc abc::print_symm { {resultArrayName {}} {paramsString {}} {expectTimeout {}}} {
	if {$resultArrayName != {}} {
		upvar $resultArrayName res
	}

	set res(success) -2
	set res(total_func_supps) null
	set res(total_struct_supps) null
	set res(total_symm) null
	set res(structural_symm) null
	set res(total_non_sym) null
	set res(total_var_pairs) null
	
    abc::command "print_symm $paramsString" res $expectTimeout

	set matched 0
	if {[regexp -- {Total func supps[[:blank:]]*=[[:blank:]]*([0-9]+)} $res(rawOutput) match v1]} {
		set res(total_func_supps) $v1
		incr matched
	}
	if {[regexp -- {Total struct supps[[:blank:]]*=[[:blank:]]*([0-9]+)} $res(rawOutput) match v1]} {
		set res(total_struct_supps) $v1
		incr matched
	}
	if {[regexp -- {Total symm[[:blank:]]*=[[:blank:]]*([0-9]+)} $res(rawOutput) match v1]} {
		set res(total_symm) $v1
		incr matched
	}
	if {[regexp -- {Structural symm[[:blank:]]*=[[:blank:]]*([0-9]+)} $res(rawOutput) match v1]} {
		set res(structural_symm) $v1
		incr matched
	}
	if {[regexp -- {Total non-sym[[:blank:]]*=[[:blank:]]*([0-9]+)} $res(rawOutput) match v1]} {
		set res(total_non_sym) $v1
		incr matched
	}
	if {[regexp -- {Total var pairs[[:blank:]]*=[[:blank:]]*([0-9]+)} $res(rawOutput) match v1]} {
		set res(total_var_pairs) $v1
		incr matched
	}


	if {[regexp -- {Empty network} $res(rawOutput)]} {
		abc::errorReturn res {ABC_E_EMPTY}
	}
    if { $matched == 0 } {
		abc::errorReturn res {ABC_E_PARSE}
    }
	return -code ok
}

# executes ABC's 'print_unate' command
# resultArrayName is an optional parameter, consists of pairs (key, value):
# keys: contains 'null' when value was not recognized:
#       shared_BDD_size [int]
#       ins             [int]
#       outs            [int]
#       glob_BDDs       [float]
#       total_supp      [int]
#       total_unate     [int]
#       unateness       [float]
#       total           [float]
#       rawOutput       [string]
#       errorCode       [string]   error message, NONE if OK
#       promptNumber    [int]      prompt number
#       cmdTime         [int]      time in milliseconds
#
# paramsString - optional string containing parameters for print_unate command,
# string is passed onto print_unate 'as is'
# return: ok or error
#
#   errors: ABC_E_EOF, ABC_E_TOUT  ABC_E_BUSY ABC_E_PARSE ABC_E_EMPTY
#
# BUG: message not detected
# This command works only for AIGs (run "strash").
proc abc::print_unate { {resultArrayName {}} {paramsString {}} {expectTimeout {}}} {

	if {$resultArrayName != {}} {
		upvar $resultArrayName res
	}

	set res(shared_BDD_size) null
	set res(ins) null
	set res(outs) null
	set res(glob_BDDs) null
	set res(total_supp) null
	set res(total_unate) null
	set res(unateness) null
	set res(total) null

    abc::command "print_unate $paramsString" res $expectTimeout
    set matched 0

	if {[regexp -- {Shared BDD size =[[:blank:]]*([0-9]+)} $res(rawOutput) match v1]} {
		set res(shared_BDD_size) $v1
		incr matched
	}
	if {[regexp -- {Ins/Outs =[[:blank:]]*([0-9]+)/[[:blank:]]*([0-9]+)} $res(rawOutput) match v1 v2]} {
		set res(ins) $v1
		set res(outs) $v2
		incr matched
	}
	if {[regexp -- {Glob BDDs =[[:blank:]]*([0-9]*.[0-9]*)} $res(rawOutput) match v1]} {
		set res(glob_BDDs) $v1
		incr matched
	}
	if {[regexp -- {Total supp =[[:blank:]]*([0-9]+)} $res(rawOutput) match v1]} {
		set res(total_supp) $v1
		incr matched
	}
	if {[regexp -- {Total unate =[[:blank:]]*([0-9]+)} $res(rawOutput) match v1]} {
		set res(total_unate) $v1
		incr matched
	}
	if {[regexp -- {Unateness =[[:blank:]]*([0-9]*.[0-9]*)} $res(rawOutput) match v1]} {
		set res(unateness) $v1
		incr matched
	}
	if {[regexp -- {Total[[:blank:]]*=[[:blank:]]*([0-9]*.[0-9]*)} $res(rawOutput) match v1]} {
		set res(total) $v1
		incr matched
	}
	if {[regexp -- {Empty network} $res(rawOutput)]} {
		abc::errorReturn res {ABC_E_EMPTY}
	}
    if { $matched == 0 } {
		abc::errorReturn res {ABC_E_PARSE}
    }
	return -code ok
}

# paramsString - optional string containing parameters for read command,
# including filename, string is passed onto read 'as is'
# resultArrayName is an optional parameter, consists of pairs (key, value)
# keys: errorCode     [string]   error message, NONE if OK
#       rawOutput     [string]   ABC's raw output
#       promptNumber  [int]      prompt number
#       cmdTime       [int]      time in milliseconds
#
#   errors: ABC_E_EOF ABC_E_TOUT  ABC_E_BUSY  ABC_E_OPEN ABC_E_PARAMS ABC_E_SUFFIX ABC_E_READ
#
proc abc::read {{resultArrayName {}}  {paramsString {}}  {expectTimeout {}}} {
	global abc::busy

	if {$resultArrayName != {}} {
		upvar $resultArrayName res
	}

    abc::command "read $paramsString" res $expectTimeout

	set regexpRes [regexp -- {Cannot open input file "([^"])*" } $res(rawOutput) match v1]
	if {$regexpRes} {
		abc::errorReturn res  {ABC_E_OPEN} $v1
	}
	set regexpRes [regexp -- {usage: read} $res(rawOutput)]
	if {$regexpRes} {
		abc::errorReturn res  {ABC_E_PARAMS}
	}
	set regexpRes [regexp -- {The generic file reader requires a known file extension} $res(rawOutput)]
	if {$regexpRes} {
		abc::errorReturn res  {ABC_E_SUFFIX}
	}
	if {[regexp -- {Reading network from file has failed} $res(rawOutput)]} {
		abc::errorReturn res  {ABC_E_READ} "" ""
	}
	return -code ok
}

# Renoding for CNF has failed.

# ERROR in Abc_NtkMiterProve(): Generated counter-example is invalid.
# Timed out after verifying %d outputs (out of %d).\n
# covered by the 'undecided' case

# tests given string and sets output values according to its content
# called from within: abc::sec, abc::ceccec, abc::dsec, abc::dprove
# parameters: buffer    [string]    string to test
# return: ABC_N_EQ_YES ABC_N_EQ_NO ABC_N_UND 
#         ABC_E_EMPTY ABC_E_PARAMS ABC_E_INCOMP
proc abc::checkEquivalence {buffer} {
	set result -1

	if {[regexp -nocase -- {not equivalent} $buffer]} {	return {ABC_N_EQ_NO} }	
	if {[regexp -nocase -- {equivalent} $buffer]}     { return {ABC_N_EQ_YES} }
	if {[regexp -nocase -- {undecided} $buffer]}      { return {ABC_N_EQ_UND} }	
	if {[regexp -nocase -- {Empty (?:current |)network} $buffer]} { return {ABC_E_EMPTY} }
	if {[regexp -nocase -- {Command line switch} $buffer]}        { return {ABC_E_PARAMS} }
	if {[regexp -nocase -- {Miter computation has failed.} $buffer]} { return {ABC_E_INCOMP} }
}

# executes ABC's 'sec' command
# resultArrayName is an optional parameter, consists of pairs (key, value):
# keys: rawOutput       [string]
#       errorCode       [string] error message, NONE if OK
#       promptNumber    [int]    prompt number
#       cmdTime         [int]    time in milliseconds
#       answer          [string] one of ABC_N_EQ_YES ABC_N_EQ_NO ABC_N_UND 
# paramsString - optional string containing parameters for sec command,
# string is passed onto sec 'as is'
#
# errors: ABC_E_EOF, ABC_E_TOUT  ABC_E_BUSY  ABC_E_EMPTY ABC_E_PARAMS ABC_E_INCOMP  ABC_E_PARSE
#
proc abc::sec { {resultArrayName {}} {paramsString {}}  {expectTimeout {}}} {
	if {$resultArrayName != {}} {
		upvar $resultArrayName res
	}
    abc::command "sec $paramsString" res $expectTimeout
	unset res(answer)
	set retCode [abc::checkEquivalence $res(rawOutput)]
	if { $retCode == {}} {
	    abc::errorReturn res  {ABC_E_PARSE}
	}
	if { [abc::extrSeverity {retCode}] == "E" } {
	    abc::errorReturn res  $retcode
	}
	set res(answer) $retCode
	return -code ok $retCode
}

# executes ABC's 'cec' command
# resultArrayName is an optional parameter, consists of pairs (key, value):
# keys: rawOutput       [string]
#       errorCode       [string] error message, NONE if OK
#       promptNumber    [int]    prompt number
#       cmdTime         [int]    time in milliseconds
#       answer          [string] one of ABC_N_EQ_YES ABC_N_EQ_NO ABC_N_UND 
# paramsString - optional string containing parameters for cec command,
#       string is passed onto cec 'as is'
# return: if no error, one of  ABC_N_EQ_YES ABC_N_EQ_NO ABC_N_UND 
#
# errors: ABC_E_EOF, ABC_E_TOUT  ABC_E_BUSY ABC_E_EMPTY ABC_E_PARAMS ABC_E_INCOMP  ABC_E_PARSE
#
proc abc::cec { {resultArrayName {}} {paramsString {}}  {expectTimeout {}}} {
	if {$resultArrayName != {}} {
		upvar $resultArrayName res
	}
    abc::command "cec $paramsString" res $expectTimeout
	#unset res(answer)
	set retCode [abc::checkEquivalence $res(rawOutput)]
	if { $retCode == {}} {
	    abc::errorReturn res  {ABC_E_PARSE}
	}
	if { [abc::extrSeverity {retCode}] == "E" } {
	    abc::errorReturn res  $retCode
	}
	set res(answer) $retCode
	return -code ok $retCode
}

# executes ABC's 'dsec' command
# resultArrayName is an optional parameter, consists of pairs (key, value):
# keys: rawOutput       [string]
#       errorCode       [string] error message, NONE if OK
#       promptNumber    [int]    prompt number
#       cmdTime         [int]    time in milliseconds
#       answer          [string] one of ABC_N_EQ_YES ABC_N_EQ_NO ABC_N_UND 
# paramsString - optional string containing parameters for dsec command,
# string is passed onto dsec 'as is'
# return: if no error, one of  ABC_N_EQ_YES ABC_N_EQ_NO ABC_N_UND 
#
# errors: ABC_E_EOF, ABC_E_TOUT  ABC_E_BUSY  ABC_E_EMPTY ABC_E_PARAMS ABC_E_INCOMP  ABC_E_PARSE
#
proc abc::dsec { {resultArrayName {}} {paramsString {}}  {expectTimeout {}}} {
	if {$resultArrayName != {}} {
		upvar $resultArrayName res
	}
    abc::command "cec $paramsString" res $expectTimeout
	unset res(answer)
	set retCode [abc::checkEquivalence $res(rawOutput)]
	if { $retCode == {}} {
	    abc::errorReturn res  {ABC_E_PARSE}
	}
	if { [abc::extrSeverity {retCode}] == "E" } {
	    abc::errorReturn res  $retCode
	}
	set res(answer) $retCode
	return -code ok $retCode
}
# executes ABC's 'dprove' command
# resultArrayName is an optional parameter, consists of pairs (key, value):
# keys: rawOutput       [string]
#       errorCode       [string] error message, NONE if OK
#       promptNumber    [int]      prompt number
#       cmdTime         [int]      time in milliseconds
#       answer          [string] one of ABC_N_EQ_YES ABC_N_EQ_NO ABC_N_UND 
# paramsString - optional string containing parameters for dprove command,
# string is passed onto dprove 'as is'
# return: if no error, one of  ABC_N_EQ_YES ABC_N_EQ_NO ABC_N_UND 
#
# errors: ABC_E_EOF, ABC_E_TOUT  ABC_E_BUSY   ABC_E_EMPTY ABC_E_PARAMS ABC_E_INCOMP  ABC_E_PARSE
#
proc abc::dprove { {resultArrayName {}} {paramsString {}}  {expectTimeout {}}} {
	if {$resultArrayName != {}} {
		upvar $resultArrayName res
	}
    abc::command "dprove $paramsString" res $expectTimeout
	unset res(answer)
	set retCode [abc::checkEquivalence $res(rawOutput)]
	if { $retCode == {}} {
	    abc::errorReturn res  {ABC_E_PARSE}
	}
	if { [abc::extrSeverity {retCode}] == "E" } {
	    abc::errorReturn res  $retCode
	}
	set res(answer) $retCode
	return -code ok $retCode
}
# BUG:
# abc: src/base/abci/abcSat.c:58: Abc_NtkMiterSat: Assertion `Abc_NtkLatchNum(pNtk) == 0' failed.
# Aborted
# NOT COVERED.
# tests given string and sets output values according to its content
# called from within: abc::iprove, abc::prove, abc::sat
# parameters: buffer    [string]    string to test
# return: ABC_N_EQ_YES ABC_N_EQ_NO ABC_N_UND 
#         ABC_E_EMPTY ABC_E_PARAMS ABC_E_INCOMP
proc abc::checkSatisfiability {buffer} {
	if {[regexp -nocase -- {satisfiable} $buffer]} { return {ABC_N_SAT_YES}}
	if {[regexp -nocase -- {unsatisfiable} $buffer]} { return {ABC_N_SAT_NO}}
	if {[regexp -nocase -- {undecided} $buffer]} {return {ABC_N_SAT_UND}}
	if {[regexp -nocase -- {Empty (?:current |)network} $buffer]} {return {ABC_E_EMPTY}}
}

# executes ABC's 'iprove' command
# resultArrayName is an optional parameter, consists of pairs (key, value):
# keys: rawOutput       [string]
#       errorCode       [string] error message, NONE if OK
#       promptNumber    [int]    prompt number
#       cmdTime         [int]    time in milliseconds
#       answer          [string] one of ABC_N_SAT_YES ABC_N_SAT_NO ABC_N_SAT_UND 
# paramsString - optional string containing parameters for iprove command,
# string is passed onto iprove 'as is'
# return: if no error, one of  ABC_N_SAT_YES ABC_N_SAT_NO ABC_N_SAT_UND 
#
# errors: ABC_E_EOF ABC_E_TOUT ABC_E_BUSY ABC_E_EMPTY ABC_E_PARSE
#
proc abc::iprove { {resultArrayName {}} {paramsString {}}  {expectTimeout {}}} {
	if {$resultArrayName != {}} {
		upvar $resultArrayName res
	}
    abc::command "iprove $paramsString" res $expectTimeout
	unset res(answer)
	set retCode [abc::checkSatisfiability $res(rawOutput)]
	if { $retCode == {}} {
	    abc::errorReturn res  {ABC_E_PARSE}
	}
	if { [abc::extrSeverity {retCode}] == "E" } {
	    abc::errorReturn res  $retCode
	}
	set res(answer) $retCode
	return -code ok $retCode
}
# executes ABC's 'prove' command
# resultArrayName is an optional parameter, consists of pairs (key, value):
# keys: rawOutput       [string]
#       errorCode       [string] error message, NONE if OK
#       promptNumber    [int]    prompt number
#       cmdTime         [int]    time in milliseconds
#       answer          [string] one of ABC_N_SAT_YES ABC_N_SAT_NO ABC_N_SAT_UND 
# paramsString - optional string containing parameters for prove command,
# string is passed onto prove 'as is'
# return: if no error, one of  ABC_N_SAT_YES ABC_N_SAT_NO ABC_N_SAT_UND 
#
# errors: ABC_E_EOF ABC_E_TOUT ABC_E_BUSY ABC_E_EMPTY ABC_E_PARSE
#
proc abc::prove { {resultArrayName {}} {paramsString {}}  {expectTimeout {}}} {
	if {$resultArrayName != {}} {
		upvar $resultArrayName res
	}
    abc::command "prove $paramsString" res $expectTimeout
	unset res(answer)
	set retCode [abc::checkSatisfiability $res(rawOutput)]
	if { $retCode == {}} {
	    abc::errorReturn res  {ABC_E_PARSE}
	}
	if { [abc::extrSeverity {retCode}] == "E" } {
	    abc::errorReturn res  $retCode
	}
	set res(answer) $retCode
	return -code ok $retCode
}
# executes ABC's 'sat' command
# resultArrayName is an optional parameter, consists of pairs (key, value):
# keys: rawOutput       [string]
#       errorCode       [string] error message, NONE if OK
#       promptNumber    [int]    prompt number
#       cmdTime         [int]    time in milliseconds
#       answer          [string] one of ABC_N_SAT_YES ABC_N_SAT_NO ABC_N_SAT_UND 
# paramsString - optional string containing parameters for sat command,
# string is passed onto sat 'as is'
# return: if no error, one of  ABC_N_SAT_YES ABC_N_SAT_NO ABC_N_SAT_UND 
#
# errors: ABC_E_EOF ABC_E_TOUT ABC_E_BUSY ABC_E_EMPTY ABC_E_PARSE
#
proc abc::sat { {resultArrayName {}} {paramsString {}}  {expectTimeout {}}} {
	if {$resultArrayName != {}} {
		upvar $resultArrayName res
	}
    abc::command "sat $paramsString" res $expectTimeout
	unset res(answer)
	set retCode [abc::checkSatisfiability $res(rawOutput)]
	if { $retCode == {}} {
	    abc::errorReturn res  {ABC_E_PARSE}
	}
	if { [abc::extrSeverity {retCode}] == "E" } {
	    abc::errorReturn res  $retCode
	}
	set res(answer) $retCode
	return -code ok $retCode
}
# resultArrayName is an optional parameter, consists of pairs (key, value):
# keys: contains 'null' when value was not recognized:
#       elapse       [float]
#       total        [float]
#       errorCode    [string] error message, NONE if OK
#       rawOutput    [string]
#       promptNumber [int]      prompt number
#       cmdTime      [int]      time in milliseconds
# paramsString - optional string containing parameters for time command,
# string is passed onto time 'as is'
# procedure return: empty
#
# errors: ABC_E_EOF ABC_E_TOUT ABC_E_BUSY ABC_E_PARSE
#
proc abc::time { {resultArrayName {}} {paramsString {}}  {expectTimeout {}}} {
	if {$resultArrayName != {}} {
		upvar $resultArrayName res
	}
	set res(elapse) null
	set res(total) null

    abc::command "time $paramsString" res $expectTimeout

	if {[regexp -- {(elapse):[[:blank:]]*([0-9]+.[0-9]+)} $res(rawOutput) match v1 v2]} {
		set res($v1) $v2
		incr matched
	}
	if {[regexp -- {(total):[[:blank:]]*([0-9]+.[0-9]+)} $res(rawOutput) match v1 v2]} {
		set res($v1) $v2
		incr matched
	}
    if { $matched == 0 } {
		abc::errorReturn res {ABC_E_PARSE}
    }
	return -code ok
}






